from __future__ import annotations

import warnings
from collections.abc import Iterable, Mapping, Sequence
from typing import Any, Optional, Protocol, Type, TypeAlias, TypeVar

from django.core.exceptions import FieldDoesNotExist
from django.db import models

AccessorContext: TypeAlias = object | Mapping[str, Any] | Sequence

# Any type that implements the protocol
AccessorType = TypeVar("AccessorType", bound="AccessorProtocol")

# An AccessorSpec allows an accessor to be specified by its string representation
AccessorSpec: TypeAlias = str | AccessorType


class AccessorProtocol(Protocol):
    """
    A string that describes how to resolve a value from an arbitrarily deeply nested object, dictionary, or sequence.
    E.g. hn.helpers.algorithms.Accessor implements this protocol

    Design note: this combines API for a generic Accessor with API for a more specific "ModelFieldAccessor"
    This violates SR principle a little, but keeps things simple and don't forsee any problems.
    """

    def __init__(self, value: str):
        """Initialise the accessor with a string specifying the "path" used to resolve access to a value."""
        ...

    def resolve(self, context: AccessorContext) -> Any:
        """Return the value of this accessor in the given context."""
        ...

    def get_field(self, model: models.Model | type[models.Model]) -> models.Field:
        """Resolve this accessor using given model as context to return the model field rather than its value."""
        ...

    def get_label(self, model: models.Model | type[models.Model]) -> str:
        """Resolve this accessor using given model as context to return the model field verbose name."""
        ...  # Note: this method is not defined on base tables2 Accessor, see extension below.


#####
# BaseAccessor - copied directly from https://github.com/jieter/django-tables2/blob/master/django_tables2/utils.py
#    to avoid otherwise unnecessary dependency on `table2`, but you should use tables2 anyhow, its awesome !!
#    Licence:  https://github.com/jieter/django-tables2/blob/master/LICENSE
#####


class BaseAccessor(str):
    """
    A string describing a path from one object to another via attribute/index
    accesses. For convenience, the class has an alias `.A` to allow for more concise code.

    Relations are separated by a ``__`` character.

    To support list-of-dicts from ``QuerySet.values()``, if the context is a dictionary,
    and the accessor is a key in the dictionary, it is returned right away.
    """

    LEGACY_SEPARATOR = "."
    SEPARATOR = "__"

    ALTERS_DATA_ERROR_FMT = "Refusing to call {method}() because `.alters_data = True`"
    LOOKUP_ERROR_FMT = "Failed lookup for key [{key}] in {context}, when resolving the accessor {accessor}"

    def __init__(self, value, callable_args=None, callable_kwargs=None):
        self.callable_args = (
            callable_args or getattr(value, "callable_args", None) or []
        )
        self.callable_kwargs = (
            callable_kwargs or getattr(value, "callable_kwargs", None) or {}
        )
        super().__init__()

    def __new__(cls, value, callable_args=None, callable_kwargs=None):
        instance = super().__new__(cls, value)
        if cls.LEGACY_SEPARATOR in value:
            instance.SEPARATOR = cls.LEGACY_SEPARATOR

            message = (
                f"Use '__' to separate path components, not '.' in accessor '{value}'"
                " (fallback will be removed in django_tables2 version 3)."
            )

            warnings.warn(message, DeprecationWarning, stacklevel=3)

        return instance

    def resolve(self, context, safe=True, quiet=False):
        """
        Return an object described by the accessor by traversing the attributes of *context*.

        Lookups are attempted in the following order:

         - dictionary (e.g. ``obj[related]``)
         - attribute (e.g. ``obj.related``)
         - list-index lookup (e.g. ``obj[int(related)]``)

        Callable objects are called, and their result is used, before
        proceeding with the resolving.

        Example::

            >>> x = Accessor("__len__")
            >>> x.resolve("brad")
            4
            >>> x = Accessor("0__upper")
            >>> x.resolve("brad")
            "B"

        If the context is a dictionary and the accessor-value is a key in it,
        the value for that key is immediately returned::

            >>> x = Accessor("user__first_name")
            >>> x.resolve({"user__first_name": "brad"})
            "brad"


        Arguments:
            context : The root/first object to traverse.
            safe (bool): Don't call anything with `alters_data = True`
            quiet (bool): Smother all exceptions and instead return `None`

        Returns:
            target object

        Raises:
            TypeError`, `AttributeError`, `KeyError`, `ValueError`
            (unless `quiet` == `True`)
        """
        # Short-circuit if the context contains a key with the exact name of the accessor,
        # supporting list-of-dicts data returned from values_list("related_model__field")
        if isinstance(context, dict) and self in context:
            return context[self]

        try:
            current = context
            for bit in self.bits:
                try:  # dictionary lookup
                    current = current[bit]
                except (TypeError, AttributeError, KeyError):
                    try:  # attribute lookup
                        current = getattr(current, bit)
                    except (TypeError, AttributeError):
                        try:  # list-index lookup
                            current = current[int(bit)]
                        except (
                            IndexError,  # list index out of range
                            ValueError,  # invalid literal for int()
                            KeyError,  # dict without `int(bit)` key
                            TypeError,  # unsubscriptable object
                        ):
                            current_context = (
                                type(current)
                                if isinstance(current, models.Model)
                                else current
                            )

                            raise ValueError(
                                self.LOOKUP_ERROR_FMT.format(
                                    key=bit, context=current_context, accessor=self
                                )
                            )
                if callable(current):
                    if safe and getattr(current, "alters_data", False):
                        raise ValueError(
                            self.ALTERS_DATA_ERROR_FMT.format(method=current.__name__)
                        )
                    if not getattr(current, "do_not_call_in_templates", False):
                        current = current(*self.callable_args, **self.callable_kwargs)
                # Important that we break in None case, or a relationship
                # spanning across a null-key will raise an exception in the
                # next iteration, instead of defaulting.
                if current is None:
                    break
            return current
        except Exception:
            if not quiet:
                raise

    @property
    def bits(self):
        if self == "":
            return ()
        return self.split(self.SEPARATOR)

    def get_field(self, model):
        """
        Return the django model field for model in context, following relations.
        """
        if not hasattr(model, "_meta"):
            return

        field = None
        for bit in self.bits:
            try:
                field = model._meta.get_field(bit)
            except FieldDoesNotExist:
                break

            if hasattr(field, "remote_field"):
                rel = getattr(field, "remote_field", None)
                model = getattr(rel, "model", model)

        return field

    def penultimate(self, context, quiet=True):
        """
        Split the accessor on the right-most separator ('__'), return a tuple with:
         - the resolved left part.
         - the remainder

        Example::

            >>> Accessor("a__b__c").penultimate({"a": {"a": 1, "b": {"c": 2, "d": 4}}})
            ({"c": 2, "d": 4}, "c")

        """
        path, _, remainder = self.rpartition(self.SEPARATOR)
        return BaseAccessor(path).resolve(context, quiet=quiet), remainder


#####
# Accessor - cornerstone building block - defines how to resolve a value from a string description.
#####
class Accessor(BaseAccessor):
    """
    A string describing a path from one object to another via attribute/index accesses.

    Relations are separated by a ``__`` character.

    Usage::

        >>> x = Accessor("__len__")
        >>> x.resolve("brad")
        4
        >>> x = Accessor("0__upper")
        >>> x.resolve("brad")
        "B"

    This class is a placeholder in case we want to eliminate dependency on tables2.
    While we have this dependency, why re-invent the wheel?
    """

    def get_label(self, model: models.Model | type[models.Model]) -> str:
        """Resolve this accessor using given model as context to return the model field verbose name."""
        if isinstance(model, models.Model):
            model = type(model)
        field = self.get_field(model)
        if field and field.verbose_name:
            return field.verbose_name
        return self.bits[-1].replace("_", " ").title()


class AccessorTransform:
    """Provides a transform from standard accessor names to prefixed accessor names"""

    def __init__(self, prefix: str = None):
        self.prefix = prefix

    def t(self, accessor: str) -> str:
        """Transform key, if it exists in this KeyMap"""
        return f"{self.prefix}__{accessor}" if self.prefix else accessor

    def __call__(self, accessors: str | dict | Iterable) -> str | dict | list:
        """Transform all value(s) in accessors and return an object of the same type"""
        match accessors:
            case str():
                return self.t(accessors)
            case dict():
                return {
                    self.t(accessor): value for accessor, value in accessors.items()
                }
            case _:
                return [self.t(accessor) for accessor in accessors]


def get_accessor(
    accessor: Optional[AccessorSpec], accessor_type: Type[AccessorType] = Accessor
) -> AccessorType | None:
    """helper: return an Accessor instance from the given spec.  Returns None if input accessor is None."""
    return accessor_type(accessor) if isinstance(accessor, str) else accessor



from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent
print(BASE_DIR)

SECRET_KEY = 'django-insecure'

DEBUG = True

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'nifty_layout'
]



from dataclasses import dataclass
from functools import partial
from typing import Any

from django.contrib.auth import get_user_model
from django.test import SimpleTestCase

from nifty_layout.components import field_labeller, DictCompositeNode, FieldNode
from nifty_layout.accessor import AccessorSpec

User = get_user_model()

class FormatterTests(SimpleTestCase):
    """Test suite for template formatter functionality."""

    def test_currency_formatter(self):
        """Test formatting numbers as currency."""
        currency_formatter = partial(self._template_formatter, template="${:,.2f}")
        self.assertEqual(currency_formatter(123.4567), "$123.46")

    @staticmethod
    def _template_formatter(value: Any, template: str, **context) -> str:
        return template.format(value, **context)


class LabellerTests(SimpleTestCase):
    """Test suite for field labeller functionality."""

    @dataclass
    class SimpleObject:
        a: int = 42
        a_label: str = "The Answer"
        b: str = "Question?"
        some_unlabelled_attr: Any = None

        def get_b_label(self):
            return f"What is {self.a_label}"

    def test_labeller_with_partial(self):
        """Test labeller usage with partials."""
        obj = self.SimpleObject()

        obj_field_labeller = partial(field_labeller, obj)
        self.assertEqual(obj_field_labeller('a'), "The Answer")
        self.assertEqual(obj_field_labeller('b'), "What is The Answer")
        self.assertEqual(obj_field_labeller('some_unlabelled_attr'), "Some Unlabelled Attr")

    def test_labeller_with_user_model(self):
        """Test labelling fields for a User instance."""
        larry = User(username="lmapmaker", first_name="Larry", last_name="Mapmaker")
        larry.get_full_name = lambda: f"{larry.first_name} {larry.last_name}"

        first_name_labeller = partial(field_labeller, accessor="first_name")
        self.assertEqual(first_name_labeller(larry), "first name")

        user_field_labeller = partial(field_labeller, larry)
        self.assertEqual(user_field_labeller("first_name"), "first name")
        self.assertEqual(user_field_labeller("last_name"), "last name")
        self.assertEqual(user_field_labeller("get_full_name"), "Get Full Name")


class CompositeNodeTests(SimpleTestCase):
    """Test suite for DictCompositeNode and related functionality."""

    def test_get_field_with_comments(self):
        """Test creating and using DictCompositeNode with field and comments."""
        checkbox = self._get_field_with_comments("my_field", "my_field_comments")

        self.assertEqual(len(checkbox), 2)
        self.assertEqual(checkbox['field'].accessor, 'my_field')
        self.assertEqual(checkbox['comments'].accessor, 'my_field_comments')
        self.assertListEqual(list(checkbox.keys()), ['field', 'comments'])

        with checkbox.as_sequence():
            self.assertListEqual(
                [fld.accessor for fld in checkbox],
                ['my_field', 'my_field_comments']
            )

    def test_binding_composite_node(self):
        """Test binding DictCompositeNode to an object."""
        obj_node = self._get_field_with_comments("a", "b")
        obj = LabellerTests.SimpleObject(a=987, a_label="A Label", b="some comments!")

        bnd_node = obj_node.bind(obj)
        self.assertEqual(bnd_node['field'].label, "A Label")
        self.assertEqual(bnd_node['field'].value, 987)
        self.assertEqual(bnd_node['comments'].value, "some comments!")

    @staticmethod
    def _get_field_with_comments(accessor: AccessorSpec, comment_accessor: AccessorSpec, **kwargs) -> DictCompositeNode:
        return DictCompositeNode(
            ("field", FieldNode(accessor=accessor, **kwargs)),
            ("comments", FieldNode(accessor=comment_accessor))
        )



"""Nifty Layout Components

A micro-framework for defining re-usable layout components from the fields of an object

- a "Layout Component" has 2 parts;
  - Component Nodes define the hierarchical structure for a sub-set of object fields.
    Each node encapsulates how to access, format, and label a field's value and its children.
    Once bound to a data object (e.g., a model instance) the "bound node" provides a
    uniform interface to access the value and label for one or more fields in an encapsulated object.
  - One or more templates that use the interface defined by a Bound Node to render the component into HTML.

- most layout components expect to work with django Models,
  but can be adapted to any object type by injecting custom accessors
"""

from __future__ import annotations

import abc
from collections.abc import Iterable, Mapping, Sequence
from contextlib import contextmanager
from functools import partial
from typing import Any, Callable, Hashable, Iterator, Optional, Type, TypeAlias

from django.db import models
from django.template import Template
from django.template.loader import get_template

from .accessor import Accessor, AccessorSpec, AccessorType, get_accessor

# A Formatter is a function that takes a value returns a formatted value
Formatter: TypeAlias = Callable[[Any, ...], str]

# A Labeller is a function that takes an accessor and returns string label for the attribute
Labeller: TypeAlias = Callable[[AccessorSpec, ...], str]


def field_labeller(obj: object | models.Model, accessor: AccessorSpec, **unused) -> str:
    """Return a label for the given accessor of given object, which could be a django Model instance.

    If `obj` is a model instance, returns the verbose name of the accessor field, if accessor is a model field.
    Otherwise, looks for magic attribute / method:  `foo_label` or `get_foo_label` on obj.
    In either case, if no labelling is provided by obj, defines a title-case label based on accessor.
    """
    accessor = Accessor(accessor) if isinstance(accessor, str) else accessor
    if isinstance(obj, models.Model):
        return accessor.get_label(obj)
    # look for "foo_label" attribute or "get_foo_label" method on object, if there isn't one, use the attribute name
    attr = accessor.bits[-1]
    label = getattr(obj, f"{attr}_label", getattr(obj, f"get_{attr}_label", None))
    if callable(label):
        label = label()
    label = label if label is not None else attr.replace("_", " ").title()
    return label


def get_formatted_labeller(
    labeller, formatter
):  # todo - add tests or change labelling/formatting logic
    def formatted_labeller(*args, **kwargs):
        return formatter(labeller(*args, **kwargs))

    return formatted_labeller


# Examples:


#########
# Component Nodes
# A Bound Node is an adapter that provides a uniform interface for access to values and labels of encapsulated object
# Its unbound Node defines access and formatting rules.
#########
#########
# Important design note: although we are defining the type hierarchy with inheritance here, the behaviours for each
#  type are assembled using composition.
#  E.g., define reusable accessor, labeller, formatter to assemble custom node types.
#  Custom node types are mainly for convenience - a commonly used assembly - think of them as syntactic sugar
#  The key here is: don't build a deeply nested inheritance hierarchy!  Use composition to customize Node behaviours.
#########


# A TemplateType can be a Template object, the string template path, or a callable that takes a single parameter.
TemplateType: TypeAlias = Template | str | Callable[["BoundNode"], Template]


class BoundNode(Iterable):
    """A node that is bound to a piece of data, usually a dict, iterable, object, or model instance
    Defines the interface for accessing a "value" and "label" from the data, and iterating / accessing its children.

    If node is a composite, all children are also bound to the same data.  Iteration is over the bound fixings nodes
    Optional template is passed through to context, and **may** be used, if parent template is so configured.
    """

    def __init__(
        self, data: Any, node: BaseNode, template: Optional[TemplateType] = None
    ):
        self.node = node
        self.data = data
        self.children = [child.bind(data) for child in node]
        self.template = self._get_template(template) if template else None

    def _get_template(self, template: TemplateType) -> Template:
        """return a template for rendering this node with."""
        match template:
            case Template():
                return template
            case str():
                return get_template(template)
            case _ if callable(template):
                return template(self)
            case _:
                raise ValueError(f"Unexpected value for template: {template}.")

    @property
    def value(self):
        """Fetch the data value from data object and format it."""
        return self.node.get_formatted_value(self.node.get_raw_value(self.data))

    @property
    def label(self):
        """Fetch the label from the data object."""
        return self.node.get_label(self.data)

    # Iterable and Sequence interface

    def __getitem__(self, index) -> BoundNode:
        """Return the bound fixings node at given 0 <= index < len(self)"""
        return self.children[index]

    def __len__(self) -> int:
        return len(self.children)

    def __iter__(self) -> Iterator[BoundNode]:
        """Iterate over bound fixings nodes."""
        return iter(self.children)


class BoundSequenceNode(BoundNode, Sequence):
    """A BoundNode representing a linear sequence of bound nodes."""

    def __init__(
        self,
        data: Any,
        node: SequenceCompositeNode,
        template: Optional[TemplateType] = None,
    ):
        """Narrowing Types only - exactly same as a BoundNode"""
        self.node = node  # only to please the type checker, which otherwise thinks node is a BaseNode
        super().__init__(data, node, template)


class BoundDictNode(BoundNode, Mapping):
    """A BoundNode representing a mapping of keys to bound nodes."""

    def __init__(
        self,
        data: Any,
        node: DictCompositeNode,
        template: Optional[TemplateType] = None,
    ):
        self.node = node  # only to please the type checker, which otherwise thinks node is a BaseNode
        with node.as_sequence():
            super().__init__(data, node, template)
        self.mapping = {k: v for k, v in zip(self.node.mapping.keys(), self.children)}

    @contextmanager
    def as_sequence(self):
        """A context manager allowing direct iteration and numeric indexing on the fixings nodes"""
        orig_class = type(self)
        try:
            self.__class__ = BoundSequenceNode
            yield self  # Hand back control to the with-block
        finally:
            self.__class__ = orig_class

    # Mapping interface
    def __iter__(self) -> Iterator[Hashable]:
        """Iterate over the map's ordered key set."""
        return iter(self.mapping)

    def __getitem__(self, key: Hashable) -> BoundNode:
        """Lookup bound fixings node in dict by key"""
        return self.mapping[key]


##############
# Unbound Nodes are a static description of the logic to access, format, and label a value.
# The intent is to define a small, declarative syntax.
# `bind` a Node to some data (and optionally a template) to get a BoundNode suitable for rendering.
#############


class BaseNode(Iterable, abc.ABC):
    """Abstract Base class for all types of Spiffy Component Nodes"""

    default_formatter: Formatter | bool = False  # default does no formatting
    default_labeller: Labeller | str | bool = False  # default has no label
    bound_node_type: type[BoundNode] = (
        BoundNode  # data type returned by `bind` operation
    )

    def __init__(
        self,
        formatter: Optional[Formatter | bool] = None,
        labeller: Optional[Labeller | str | bool] = None,
        template: Optional[TemplateType] = None,
    ):
        """Parameters passed as None default to Node type's default values."""
        self.formatter = (
            formatter if formatter is not None else type(self).default_formatter
        )
        self.labeller = (
            labeller if labeller is not None else type(self).default_labeller
        )
        self.template = template

    def bind(self, data: Any, template: Optional[TemplateType] = None) -> BoundNode:
        """Return a BoundNode that binds the given data (and template) to this node"""
        template = template or self.template
        return self.bound_node_type(data=data, node=self, template=template)

    # Value / Label extraction API

    def get_raw_value(self, data: Any) -> Any:
        """Extract and return the raw value for this Node from the given data object"""
        return data

    def get_formatted_value(self, value: Any) -> str | Any:
        """Return formatted version of given raw value"""
        return self.formatter(value) if self.formatter else value

    def get_label(self, data: Any) -> str | None:
        """Return a label for the given data object"""
        return (
            self.labeller(data)
            if callable(self.labeller)
            else str(self.labeller) if self.labeller else None
        )

    # Iterable interface

    def __iter__(self):
        return iter([])


# Concrete Node types fall in 2 basic classes: Atomic and Composite


class AtomicNode(BaseNode, abc.ABC):
    """Abstract Base class for an atomic Components (no children)"""

    pass


class FieldNode(AtomicNode):
    """Basic atomic node for encapsulating access to the data for a single field using a Accessor"""

    default_accessor_type: Type[AccessorType] = (
        Accessor  # The Accessor type used to wrap naked accessor spec.
    )
    default_labeller: Labeller | str | bool = field_labeller

    def __init__(
        self,
        accessor: AccessorSpec,
        formatter: Optional[Formatter | bool] = None,
        labeller: Optional[Labeller | str | bool] = None,
        template: Optional[TemplateType] = None,
    ):
        """Parameters passed as None default to Node type's default values."""
        labeller = (
            labeller
            if labeller is not None
            else partial(type(self).default_labeller, accessor=accessor)
        )
        super().__init__(formatter=formatter, labeller=labeller, template=template)
        self.accessor = get_accessor(accessor, type(self).default_accessor_type)

    def get_raw_value(self, data: Any) -> Any:
        """Extract and return the raw value for this Node from the given data object"""
        return self.accessor.resolve(data)


# define extended FieldNodes for common use-cases, assembled with custom labellers and formatters
# E.g., CurrencyFieldNode, DecimalFieldNode, DateField, DateTimeField,...


class CompositeNode(BaseNode, abc.ABC):
    """Abstract Base class for a composite iterable component with zero or more ordered children"""

    default_child_node_type: type[FieldNode] = (
        FieldNode  # "naked" children are wrapped in this Node type.
    )

    def __init__(
        self,
        *children: BaseNode | AccessorSpec,
        formatter: Optional[
            Formatter | bool
        ] = None,  # default: use class default_formatter
        labeller: Optional[
            Labeller | str | bool
        ] = None,  # default: use class default_labeller
        child_node_type: Optional[type[BaseNode]] = None,
        **attributes: str,
    ):
        super().__init__(formatter=formatter, labeller=labeller, **attributes)
        self.child_node_type = child_node_type or type(self).default_child_node_type
        self.children = [self.wrap_child(child) for child in children]

    def wrap_child(self, child: BaseNode | AccessorSpec) -> FieldNode:
        return child if isinstance(child, BaseNode) else self.child_node_type(child)

    # Iterable interface and Sequence interface
    def __iter__(self):
        return iter(self.children)

    def __getitem__(self, index: int) -> FieldNode:
        return self.children[index]

    def __len__(self) -> int:
        return len(self.children)


class SequenceCompositeNode(CompositeNode, Sequence):
    """A Composite node that behaves as an iterable of ordered fixings nodes"""

    bound_node_type = BoundSequenceNode


# A MappingElement can be defined by any dict-like object or by a 2-tuple (key, value)
# When used to initialize a DictCompositeNode, all MappingElements are combined, in order, to form a single dict.
MappingElement: TypeAlias = (
    Mapping[Hashable, BaseNode | AccessorSpec]
    | tuple[Hashable, BaseNode | AccessorSpec]
)


class DictCompositeNode(CompositeNode, Mapping):
    """A Composite that allows fixings nodes to be looked up with a key. Caution: iteration is over keys not values!"""

    bound_node_type = BoundDictNode

    def __init__(self, *children: MappingElement, **kwargs):
        """Children can be looked up by key, and iteration is over keys.  dict semantics."""
        mappings = (
            child if hasattr(child, "items") else dict((child,)) for child in children
        )
        mapping = {k: v for mapping in mappings for k, v in mapping.items()}

        super().__init__(*mapping.values(), **kwargs)
        assert len(mapping.keys()) == len(self.children)
        self.mapping = {k: v for k, v in zip(mapping.keys(), self.children)}

    @contextmanager
    def as_sequence(self):
        """A context manager allowing direct iteration and numeric indexing on the fixings nodes"""
        orig_class = type(self)
        try:
            self.__class__ = SequenceCompositeNode
            yield self  # Hand back control to the with-block
        finally:
            self.__class__ = orig_class

    # Mapping interface
    def __iter__(self) -> Iterator[Hashable]:
        """Iterate over the map's ordered key set."""
        return iter(self.mapping)

    def __getitem__(self, key: Hashable) -> FieldNode:
        """Lookup fixings node in map by key"""
        return self.mapping[key]


class Seq(SequenceCompositeNode):

    def __init__(self, *children: BaseNode | AccessorSpec | Iterable, **kwargs):
        if len(children) == 1 and isinstance(children[0], (list, tuple, set)):
            # Instead of wrapping the iterable itself in `child_node_class`, we wrap each element individually
            # enables us to use `child_node_class=Seq` to create a nested sequence
            children = children[0]
        super().__init__(*children, **kwargs)
